from __future__ import annotations
from typing import Dict, List

class Metadata:
    def __init__(self, predecessors:List[str], fragment_id:str=None, output_channel:str=None, custom:Dict[str, str]=None):
        self.output_channel = output_channel
        self.custom = custom
        self.fragment_id = fragment_id
        self.predecessors = predecessors
    
    def __str__(self):
        return str(self.__dict__)

    def to_json(self) -> dict:
        return self.__dict__

    @classmethod
    def from_json(cls, d:dict) -> cls:
        if not isinstance(d, dict):
            raise TypeError("Argument 'd' is not of type 'dict'")
        result = cls([])
        for key in d.keys():
            result.__setattr__(key, d[key])

        return result


from __future__ import annotations

class KillMessage:
    def __init__(self):
        self.status = "complete"
    
    def to_json(self) -> dict:
        return self.__dict__

    @classmethod
    def from_json(cls, d:dict) -> cls:
        if not isinstance(d, dict):
            raise TypeError("Argument 'd' is not of type 'dict'")
        result = cls()
        for key in result.__dict__.keys():
            result.__setattr__(key, d[key])
        return result


from __future__ import annotations

class FinishedFragmentMessage:
    def __init__(self, fragment_id:str):
        self.fragment_id = fragment_id
    
    def to_json(self) -> dict:
        return self.__dict__

    @classmethod
    def from_json(cls, d:dict) -> cls:
        if not isinstance(d, dict):
            raise TypeError("Argument 'd' is not of type 'dict'")
        result = cls("")
        for key in result.__dict__.keys():
            result.__setattr__(key, d[key])
        return result


from __future__ import annotations
from typing import List

from pyterum.metadata import Metadata

class LocalFileDesc:
    def __init__(self, name:str, path:str):
        self.name = name
        self.path = path

    def __str__(self):
        return "{ name: " + self.name + ", path: " + self.path + " }"

    def to_json(self) -> dict:
        return self.__dict__

    @classmethod
    def from_json(cls, d:dict) -> cls:
        if not isinstance(d, dict):
            raise TypeError("Argument 'd' is not of type 'dict'")
        result = cls("", "")
        for key in result.__dict__.keys():
            result.__setattr__(key, d[key])
        return result


class LocalFragmentDesc:
    def __init__(self, files:List[LocalFileDesc], predecessors:List[str]):
        self.files = files
        self.metadata = Metadata(predecessors=predecessors)
    
    def __str__(self):
        result = "{\n"
        result += "" if self.metadata == None else "    metadata: " + str(self.metadata) + "\n"
        result += "    files: ["
        for f in self.files:
            result += "\n        " + str(f)
        result += ("" if len(self.files) == 0 else "\n    ") + "]\n"
        result += "}"
        return result

    def to_json(self) -> dict:
        result = {}
        for key in self.__dict__.keys():
            if self.__dict__[key] != None:
                result[key] = self.__dict__[key]
                
        result["files"] = [f.to_json() for f in self.files]
        result["metadata"] = self.metadata.to_json()
        return result

    @classmethod
    def from_json(cls, d:dict) -> cls:
        if not isinstance(d, dict):
            raise TypeError("Argument 'd' is not of type 'dict'")

        result = cls([], [])
        result.files = [LocalFileDesc.from_json(f) for f in d["files"]]
        result.metadata = Metadata.from_json(d["metadata"])

        return result

    def add_predecessor(self, predecessor:LocalFragmentDesc):
        self.metadata.predecessors.append(predecessor.metadata.fragment_id)

    def add_file(self, filedesc:LocalFileDesc):
        self.files.append(filedesc)


from __future__ import annotations
from typing import List

from pyterum.local_fragment_desc import LocalFileDesc


class FragmenterInputMessage:
    def __init__(self, data_files:List[str], config_files:List[LocalFileDesc]):
        self.config_files = config_files
        self.data_files = data_files
    
    def __str__(self):
        return str(json.dumps(self))

    def to_json(self) -> dict:
        result = {}
        result["data_files"] = [f.to_json() for f in self.files]
        result["config_files"] = self.config_files
        return result

    @classmethod
    def from_json(cls, d:dict) -> cls:
        if not isinstance(d, dict):
            raise TypeError("Argument 'd' is not of type 'dict'")

        result = cls([], [])
        result.config_files = [LocalFileDesc.from_json(f) for f in d["config_files"]]
        result.data_files = d["config_files"]

        return result


from typing import List

from pyterum.socket_conn import SocketConn
from pyterum.kill_message import KillMessage
from pyterum.fragmenter_input_message import FragmenterInputMessage
from pyterum import env
from pyterum.logger import logger

class FragmenterInput(SocketConn):

    def __init__(self, address:str=None):
        if address == None:
            env.verify_fragmenter_envs()
            address = env.FRAGMENTER_INPUT
        super().__init__(address, retry_policy={"consume": 0})

        logger.info(f"Initializing FragmenterInput...")
        self.connect()
        self.produce = None
        self._consumer = super().consumer()

    # Yields None if the message is the kill message, indicating that there will be no more messages
    def consumer(self) -> List[str]:
        while True:
            msg = next(self._consumer)
            output = None
            try:
                output = FragmenterInputMessage.from_json(msg)
            except (KeyError, TypeError) as errFrag:
                try:
                    KillMessage.from_json(msg)
                except (KeyError, TypeError) as errKill:
                    logger.debug(errFrag)
                    logger.debug(errKill)
                    raise Exception("Could not parse message as FragmenterInputMessage nor as KillMessage")
            yield output

class FragmenterOutput(SocketConn):

    def __init__(self, address:str=None):
        if address == None:
            env.verify_fragmenter_envs()
            address = env.FRAGMENTER_OUTPUT
        super().__init__(address, retry_policy={"produce": 0})

        logger.info(f"Initializing FragmenterOutput...")
        self.connect()
        self.consumer = None

    def produce(self, data:List[str]):
        super().produce(data)

    # To send that the fragmenter is done
    def produce_kill(self):
        super().produce(KillMessage().to_json())


from socket import socket
import struct
import json

from pyterum import env

def _encode_msg_size(size: int) -> bytes:
    return struct.pack("<I", size)

def _decode_msg_size(size_bytes: bytes) -> int:
    return struct.unpack("<I", size_bytes)[0]

def _encode_bytes(content: bytes) -> bytes:
    size = len(content)
    return _encode_msg_size(size) + content

def _decode_msg(target: socket):
    enc_msg_size = target.recv(env.ENC_MSG_SIZE_LENGTH)
    msg_size = _decode_msg_size(enc_msg_size)
    enc_msg = target.recv(msg_size)
    return json.loads(enc_msg)

def _encode_msg(content) -> bytes:
    return _encode_bytes(json.dumps(content).encode("utf-8"))


def send_to(sock:socket, data):
    encoded = _encode_msg(data)
    sock.send(encoded)

def receive_from(sock:socket):
    return _decode_msg(sock)


import os

# Location of the shared volume between sidecar and the processing step
DATA_VOLUME_PATH = os.getenv("DATA_VOLUME_PATH", "")

# Socket file paths for both what comes in and goes out
FRAGMENTER_INPUT = DATA_VOLUME_PATH + "/" + os.getenv("FRAGMENTER_INPUT", default="")
FRAGMENTER_OUTPUT = DATA_VOLUME_PATH + "/" + os.getenv("FRAGMENTER_OUTPUT", default="")

# Socket file paths for both what comes in and goes out 
TRANSFORMATION_STEP_INPUT = DATA_VOLUME_PATH + "/" + os.getenv("TRANSFORMATION_STEP_INPUT", default="")
TRANSFORMATION_STEP_OUTPUT = DATA_VOLUME_PATH + "/" + os.getenv("TRANSFORMATION_STEP_OUTPUT", default="")

# Size encoding of message with a default value of 4
ENC_MSG_SIZE_LENGTH = os.getenv("ENC_MSG_SIZE_LENGTH", default=4)

EXAMPLE_SOCKET_INPUT = "./pyterum_example_sockets/example_in.sock"
EXAMPLE_SOCKET_OUTPUT = "./pyterum_example_sockets/example_out.sock"

def verify_fragmenter_envs():
    assert DATA_VOLUME_PATH != ""
    assert FRAGMENTER_INPUT.endswith(".sock")
    assert FRAGMENTER_OUTPUT.endswith(".sock")

def verify_transformation_step_envs():    
    assert DATA_VOLUME_PATH != ""
    assert TRANSFORMATION_STEP_INPUT.endswith(".sock")
    assert TRANSFORMATION_STEP_OUTPUT.endswith(".sock")



import logging

logger = logging.getLogger('pyterum')
logger.setLevel(logging.DEBUG)

# create file handler which logs even debug messages
fh = logging.FileHandler('pyterum_debug.log')
fh.setLevel(logging.DEBUG)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)

# create formatter and add it to the handlers
formatter = logging.Formatter('%(asctime)s: %(funcName)s [%(levelname)s]  %(message)s')
fh.setFormatter(formatter)
ch.setFormatter(formatter)

# add the handlers to the logger
logger.addHandler(fh)
logger.addHandler(ch)


from typing import List

from pyterum.socket_conn import SocketConn
from pyterum.local_fragment_desc import LocalFragmentDesc
from pyterum.kill_message import KillMessage
from pyterum.finished_fragment_message import FinishedFragmentMessage
from pyterum import env
from pyterum.logger import logger

class TransformationStepInput(SocketConn):

    def __init__(self, address:str=None):
        if address == None:
            env.verify_transformation_step_envs()
            address = env.TRANSFORMATION_STEP_INPUT
        super().__init__(address, retry_policy={"consume": -1})

        logger.info(f"Initializing TransformationStepInput...")
        self.connect()
        self.produce = None
        self._consumer = super().consumer()

    # Yields None if the message is the kill message, indicating that there will be no more messages
    def consumer(self) -> LocalFragmentDesc:
        while True:
            msg = next(self._consumer)
            output = None
            try:
                output = LocalFragmentDesc.from_json(msg)
            except (KeyError, TypeError) as errFrag:
                try:
                    KillMessage.from_json(msg)
                except (KeyError, TypeError) as errKill:
                    logger.debug(errFrag)
                    logger.debug(errKill)
                    raise Exception("Could not parse message as LocalFragmentDesc nor as KillMessage")
                
            yield output


class TransformationStepOutput(SocketConn):

    def __init__(self, address:str=None):
        if address == None:
            env.verify_transformation_step_envs()
            address = env.TRANSFORMATION_STEP_OUTPUT
        super().__init__(address, retry_policy={"produce": -1})

        logger.info(f"Initializing TransformationStepOutput...")
        self.connect()
        self.consumer = None

    def produce(self, data:LocalFragmentDesc):
        super().produce(data.to_json())

    # To send that the fragmenter has completed
    def produce_kill(self):
        super().produce(KillMessage().to_json())

    def done_with(self, fragment:LocalFragmentDesc):
        super().produce(FinishedFragmentMessage(fragment.metadata.fragment_id).to_json())



import time
import sys
from socket import socket, AF_UNIX, SHUT_RDWR

from pyterum import transmit
from pyterum.logger import logger

class SocketConn:
    def __init__(self, address:str, retry_policy:dict={str:int}):
        self.address:str = address
        self.socket:socket = None
        self.retry_policy = {
            "connect": -1 if "connect" not in retry_policy else retry_policy["connect"], # Initial connection
            "consume": 0 if "consume" not in retry_policy else retry_policy["consume"],
            "produce": 0 if "produce" not in retry_policy else retry_policy["produce"] 
        }
        self.retry_interval = 5

    # Retry infinitely if retries < 0
    # Raises an error if it cannot connect within the amount of retries
    def _connect(self, retries:int):
        while True:
            try:
                self.socket = socket(AF_UNIX)
                self.socket.connect(self.address)
                return
            except Exception as err:
                if retries == 0:
                    raise err
                retries_str = " " if retries < 0 else str(retries)+ " more time(s) "
                logger.warn(f"Couldn't reach host at {self.address}, retrying {retries_str}in {self.retry_interval} seconds...")
                time.sleep(self.retry_interval)

            if retries > 0:
                retries -= 1

    def connect(self):
        self._connect(self.retry_policy["connect"])

    def close(self):
        try:
            self.socket.shutdown(SHUT_RDWR)
            self.socket.close()
        except Exception as err:
            logger.warn(f"Could not close connection due to '{err}'")

    def _reconnect(self, retries:int):
        self.close()
        self._connect(retries)

    def _silent_reconnect(self, retries:int):
        try:
            self.close()
            self._connect(retries)
        except Exception:
            pass

    def reconnect(self):
        self.close()
        self.connect()

    def consumer(self):
        retries = self.retry_policy["consume"]
        retrying = False
        while True:
            try:
                # Decode messages send over this socket one-by-one
                msg = transmit.receive_from(self.socket)
                retrying = False
                yield msg
            except Exception as err:
                if not retrying:
                    logger.warn(f"Could not consume message due to '{err}'")

                if abs(retries) > 0:
                    retrying = True
                    retries -= 1 if retries > 0 else 0
                    logger.info(f"Retrying consume after attempting to reconnect")
                    self._silent_reconnect(0)
                    time.sleep(self.retry_interval)
                else:
                    logger.error(f"Failed to consume message")
                    raise err
        self.close()

    def produce(self, data):
        retries = self.retry_policy["produce"]
        retrying = False
        while True:
            try:
                transmit.send_to(self.socket, data)
                retrying = False
                break
            except Exception as err:
                if not retrying:
                    logger.warn(f"Could not produce message due to '{err}'")

                if abs(retries) > 0:
                    retrying = True
                    retries -= 1 if retries > 0 else 0
                    logger.info(f"Retrying produce of '{data}' after attempting to reconnect")
                    self._silent_reconnect(0)
                    time.sleep(self.retry_interval)
                else:
                    logger.error(f"Failed to produce message")
                    raise err



from pyterum import env
from pyterum.transformation_step import TransformationStepInput, TransformationStepOutput
from pyterum.fragmenter import FragmenterInput, FragmenterOutput 
from pyterum.local_fragment_desc import LocalFileDesc, LocalFragmentDesc
from pyterum.metadata import Metadata